import asyncio
import inspect
from pathlib import Path
from typing import Optional, TypedDict, Union

from pyrogram import Client, enums, raw
from pyrogram.connection import Connection
from pyrogram.connection.transport import TCP
from pyrogram.storage import Storage


class ClientArgTypes(TypedDict, total=False):
	api_id: Union[int, str, None]
	api_hash: Union[str, None]
	app_version: str
	device_model: str
	system_version: str
	lang_pack: str
	lang_code: str
	system_lang_code: str
	ipv6: Union[bool, None]
	proxy: Union[dict, None]
	test_mode: Union[bool, None]
	bot_token: Union[str, None]
	session_string: Union[str, None]
	in_memory: Union[bool, None]
	phone_number: Union[str, None]
	phone_code: Union[str, None]
	password: Union[str, None]
	workers: int
	workdir: Union[str, Path]
	plugins: Union[dict, None]
	parse_mode: "enums.ParseMode"
	no_updates: Union[bool, None]
	skip_updates: Union[bool, None]
	takeout: Union[bool, None]
	sleep_threshold: int
	hide_password: Union[bool, None]
	max_concurrent_transmissions: int
	max_message_cache_size: int
	max_topic_cache_size: int
	storage_engine: Union[Storage, None]
	client_platform: "enums.ClientPlatform"
	fetch_replies: Union[bool, None]
	fetch_topics: Union[bool, None]
	fetch_stories: Union[bool, None]
	init_connection_params: Optional["raw.base.JSONValue"]
	connection_factory: type[Connection]
	protocol_factory: type[TCP]
	loop: Union[asyncio.AbstractEventLoop, None]

__fullargspec = inspect.getfullargspec(Client)
if __fullargspec.defaults is None:
    # pyromod patched the client and we cannot access it
    __fullargspec = inspect.getfullargspec(Client.old__init__)
__diff: int = len(__fullargspec.args) - len(__fullargspec.defaults)
default_args: ClientArgTypes = {arg[0]: arg[1] for arg in zip(__fullargspec.args[__diff:], __fullargspec.defaults)}



from collections.abc import Callable
from typing import Union

import pyrogram
from pyrogram.filters import Filter

import pypoligram
from pypoligram.filters import ALL
from pypoligram.filters import Filter as PFilter


class OnMessageReaction:
	def on_message_reaction(
		self: Union["OnMessageReaction", PFilter, Filter, None] = None,
		client_filters: Union[PFilter, Filter, None] = None,
		filters: Union[Filter, None] = None,
		group: int = 0
	) -> Callable:
		"""Decorator for handling message reactions.
  
		This does the same thing as :meth:`~pypoligram.ClientManager.add_handler` using the
		:obj:`~pyrogram.handlers.MessageReactionHandler`.
  
		Parameters:
			client_filters (:obj:`~pypoligram.filters`, *optional*):
				Pass one or more filters to allow only a subset of clients to be passed
				in your function.

			filters (:obj:`~pyrogram.filters`, *optional*):
				Pass one or more filters to allow only a subset of message reactions to be passed
				in your function.
	
			group (``int``, *optional*):
				The group identifier, defaults to 0.
		"""

		def decorator(func: Callable) -> Callable:
			nonlocal self, client_filters, filters, group
			if isinstance(self, pypoligram.ClientManager):
				self.add_handler(pyrogram.handlers.MessageReactionHandler(func, filters), client_filters or ALL, group)
			elif isinstance(self, Union[PFilter, Filter]) or self is None:
				if not hasattr(func, "handlers"):
					func.handlers = []
				if isinstance(self, PFilter):
					client_filters, self = self, client_filters
				if isinstance(self, Filter):
					filters, self = self, filters
				if isinstance(self, int):
					group = self or 0

				func.handlers.append(
					(
						pyrogram.handlers.MessageReactionHandler(func, filters),
						client_filters or ALL,
						group
					)
				)

			return func

		return decorator



from collections.abc import Callable
from typing import Union

import pyrogram
from pyrogram.filters import Filter, create

import pypoligram


class Filter:
	def __call__(self, clients: "pypoligram.ClientManager", client: "pyrogram.Client"):
		raise NotImplementedError

	def __invert__(self):
		return InvertFilter(self)

	def __and__(self, other):
		return AndFilter(self, other)

	def __or__(self, other):
		return OrFilter(self, other)

	def check(self, clients: "pypoligram.ClientManager", client: "pyrogram.Client"):
		return self(clients, client)


class InvertFilter(Filter):
	def __init__(self, base):
		self.base = base

	def __call__(self, clients: "pypoligram.ClientManager", client: "pyrogram.Client"):
		return not self.base(clients, client)


class AndFilter(Filter):
	def __init__(self, base, other):
		self.base = base
		self.other = other

	def __call__(self, clients: "pypoligram.ClientManager", client: "pyrogram.Client"):
		if not self.base(clients, client):
			return False
		return self.other(clients, client)


class OrFilter(Filter):
	def __init__(self, base, other):
		self.base = base
		self.other = other

	def __call__(self, clients: "pypoligram.ClientManager", client: "pyrogram.Client"):
		if self.base(clients, client):
			return True
		return self.other(clients, client)


CUSTOM_FILTER_NAME = "CustomFilter"


def create(func: Callable, name: str = None, **kwargs) -> Filter:
	"""Easy way to create a custom filter
 
	Custom filters give you extra flexibility to filter clients based on your own logic.
 
	Parameters:	
		func (``Callable``):
			A function that takes three positional arguments *(filter, manager, client)* and returns a boolean: True if the
			client should be registered the handler, False otherwise.
			The *filter* argument refers to filter itself and can be used to access keyword arguments (read below).
			The *manager* argument refers to the :obj:`pypoligram.ClientManager` and can be used to access the clients.
			The *client* argument refers to the :obj:`pyrogram.Client` and can be used to access the client.
   
		name (``str``, *optional*):
			The name of the filter. If not provided, the name of the function will be used.
			Defaults to "CustomFilter".
   
		**kwargs (``dict``, *optional*):
			Any additional keyword arguments that you want to pass to the filter.
			These will be available in the *filter* argument of the function.
	"""
	return type(
		name or func.__name__ or CUSTOM_FILTER_NAME,
		(Filter,),
		{"__call__": func, **kwargs}
	)()


# region all_filter
def all_filter(_, __, ___):
	return True


ALL = create(all_filter, "ALL")
"""Filter all clients."""


# endregion

# region user_filter
def user_filter(_, __, client: "pyrogram.Client"):
	return not bool(client.bot_token)


USER = create(user_filter, "USER")
"""Filter all users (non-bot clients).
This filter detects if the client is a user or not by checking the *bot_token* attribute."""


# endregion

# region bot_filter
def bot_filter(_, __, client: "pyrogram.Client"):
	return bool(client.bot_token)


BOT = create(bot_filter, "BOT")
"""Filter all bots (bot clients).
This filter detects if the client is a bot or not by checking the *bot_token* attribute."""


# endregion

# region client_filter
class client(Filter, set):
	"""Filter clients by its *name* attribute.
 
	Parameters:
		clients (``str`` | ``list``):
			A string or a list of strings containing the names of the clients to filter.
   	"""

	def __init__(self, clients: Union[str, list[str]]):
		clients = [] if clients is None else clients if isinstance(clients, list) else [clients]
		super().__init__(clients)

	def __call__(self, _, client: "pyrogram.Client"):
		return client.name in self

# endregion



from collections.abc import Callable
from typing import Union

import pyrogram

import pypoligram
from pypoligram.filters import ALL
from pypoligram.filters import Filter as PFilter


class OnDisconnect:
	def on_disconnect(self: Union["OnDisconnect", PFilter, None] = None, client_filters: Union[PFilter, None] = None):
		"""Decorator for handling disconnections.
  
		This does the same thing as :meth:`~pypoligram.ClientManager.add_handler` using the
		:obj:`~pyrogram.handlers.DisconnectHandler`.
  
		Parameters:
			client_filters (:obj:`~pypoligram.filters`, *optional*):
				Pass one or more filters to allow only a subset of clients to be passed
				in your function.
		"""

		def decorator(func: Callable) -> Callable:
			nonlocal self, client_filters
			if isinstance(self, pypoligram.ClientManager):
				self.add_handler(pyrogram.handlers.DisconnectHandler(func), client_filters or ALL)
			else:
				if not hasattr(func, "handlers"):
					func.handlers = []
				if isinstance(self, PFilter) or self is None:
					client_filters, self = self, client_filters

				func.handlers.append((pyrogram.handlers.DisconnectHandler(func), client_filters or ALL, 0))

			return func

		return decorator



import inspect

try:
	from pyrogram.sync import async_to_sync  #type: ignore
except ImportError:
	def async_to_sync(source, name):
		"""if we can't import it, we will just pass it.
		"""

from .clients import ClientManager


def wrap(source):
	for name in dir(source):
		method = getattr(source, name)

		if not name.startswith("_"):
			if inspect.iscoroutinefunction(method) or inspect.isasyncgenfunction(method):
				async_to_sync(source, name)

wrap(ClientManager)



__version__ = "0.3.0"
__license__ = "GNU Lesser General Public License v3.0 (LGPL-3.0)"
__copyright__ = "Copyright (C) 2023-present, MemoKing34 <https://github.com/MemoKing34>"


from . import filters, sync
from .clients import ClientManager, idle



from .on_callback_query import OnCallbackQuery
from .on_chat_boost import OnChatBoost
from .on_chat_join_request import OnChatJoinRequest
from .on_chat_member_updated import OnChatMemberUpdated
from .on_chosen_inline_result import OnChosenInlineResult
from .on_deleted_messages import OnDeletedMessages
from .on_disconnect import OnDisconnect
from .on_edited_message import OnEditedMessage
from .on_inline_query import OnInlineQuery
from .on_message import OnMessage
from .on_message_reaction import OnMessageReaction
from .on_message_reaction_count import OnMessageReactionCount
from .on_poll import OnPoll
from .on_pre_checkout_query import OnPreCheckoutQuery
from .on_purchased_paid_media import OnPurchasedPaidMedia
from .on_raw_update import OnRawUpdate
from .on_shipping_query import OnShippingQuery
from .on_story import OnStory
from .on_user_status import OnUserStatus


class Decorators(
	OnCallbackQuery,
	OnChatBoost,
	OnChatJoinRequest,
	OnChatMemberUpdated,
	OnChosenInlineResult,
	OnDeletedMessages,
	OnDisconnect,
	OnEditedMessage,
	OnInlineQuery,
	OnMessage,
	OnMessageReaction,
	OnMessageReactionCount,
	OnPoll,
	OnPreCheckoutQuery,
	OnPurchasedPaidMedia,
	OnRawUpdate,
	OnShippingQuery,
	OnStory,
	OnUserStatus,
):
	pass



import asyncio
import inspect
import logging
from collections.abc import Iterable
from concurrent.futures.thread import ThreadPoolExecutor
from importlib import import_module
from pathlib import Path
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from pyrogram import Client, compose, idle
from pyrogram.dispatcher import Dispatcher
from pyrogram.handlers.handler import Handler

from .arg_types import ClientArgTypes, default_args
from .decorators import Decorators
from .dispatcher import Dispatcher as PDispatcher
from .filters import ALL as pALL
from .filters import Filter

log = logging.getLogger(__name__)


class ClientManager(Decorators):
	"""Pypoligram's Client Manager, the main class to manage multiple clients.
 
	Parameters:
		clients (``iterable``, *optional*):
			An iterable of :obj:`pyrogram.Client` instances to be added to the manager.
			You can pass a list, tuple, set, any other iterable or nothing at all.
			Defaults to ``None``.
	
		name (``str``, *optional*):
			The name of the manager. Defaults to ``"Clients"``.
	
		plugins (``dict``, *optional*):
			Smart Plugins settings as dict, e.g.: *dict(root="plugins")*.
			This plugin has some extra settings from normal pyrogram plugin settings in ordor to apply
			all the clients. For more information, please refer to the documentation.
   			Defaults to ``None``.
	
		dont_modify (``bool``, *optional*):
			If set to ``True``, the clients will not be modified. That means handler functions will not be
			get a manager argument and the dispatcher will not be replaced.
			You can use this option to add a client to the manager and then use it as a normal client.
   			Defaults to ``False``.
	
		kwargs (``dict``, *optional*):
			A dictionary with the kwargs to be passed to the all clients. If client has a different value for the
			argument, it will be ignored. You can use this option to set the same value for all clients.
   			Defaults to ``None``.
	"""

	def __init__(
     	self,
      	clients: Iterable[Client] = None, *,
       	name: str = "Clients",
		plugins: dict = None,
  		dont_modify: bool = False,
		kwargs: ClientArgTypes = None
  	):
		if kwargs is None:
			kwargs = {}
		self._clients: set[Client] = set()
		self.name = str(name)
		self.executor = ThreadPoolExecutor(4, thread_name_prefix="Clients")
		self.loop = asyncio.get_event_loop()
		self.plugins = plugins
		self._kwargs = kwargs.copy()
		self.dont_modify = dont_modify
		if clients is not None:
			[self.add_client(client) for client in clients]

	def add_client(self, client: Client, dont_add_kwargs: bool = False) -> Client:
		"""Add a client to the manager.
  
		Parameters:
			client (:obj:`pyrogram.Client`):
				The client to be added.
	
			dont_add_kwargs (``bool``, *optional*):
				Whether to add the kwargs to the client or not. Defaults to False.
    
		Returns:
			:obj:`~pyrogram.Client`: The added client itself.
   
		Example:
			.. code-block:: python
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager()
	
				client1 = Client("my_account1")
				client2 = Client("my_account2")
	
				manager.add_client(client1)
				manager.add_client(client2)
		"""
		self._clients.add(client)
		if not self.dont_modify:
			client.dispatcher = PDispatcher(client, self)
		client.plugins = None if self.plugins else client.plugins
		client._clients = self
		if not dont_add_kwargs:
			if self._kwargs:
				for name, value in self._kwargs.items():
					c_value = getattr(client, name)
					if c_value == default_args.get(name):  # type: ignore
						setattr(client, name, value)
		return client

	def discard_client(self, client: Client) -> Client:
		"""Remove a client from the manager.

		You have to give the exact same client instance you added to the manager to remove it. You can use the return 
		value of :metm:`~pypoligram.ClientManager.add_client` method, the client itself, and pass it directly. If the 
		client was not added to the manager, it will be ignored.
  
		Parameters:
			client (:obj:`pyrogram.Client`):
				The client to be removed.
    
		Returns:
			:obj:`~pyrogram.Client`: The removed client itself.
   
		Example:
			.. code-block:: python
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager()
	
				client1 = Client("my_account1")
				client2 = Client("my_account2")
	
				manager.add_client(client1)
				manager.add_client(client2)
	
				manager.discard_client(client1)
		"""
		if not self.dont_modify:
			client.dispatcher = Dispatcher(client)
		self._clients.discard(client)
		try:
			delattr(client, "_clients")
		except AttributeError:
			pass
		return client

	def add_handler(self, handler: Handler, filters: Filter = pALL, group: int = 0, **kwargs) -> tuple[Handler, int]:
		"""Register an update handler to multiple clients.
  
		You can use it just like :meth:`pyrogram.Client.add_handler` but it will register the handler to all clients in
  		the manager that pass :obj:`pypoligram.filters`.
    
		Parameters:
			handler (``Handler``):
				The handler to be registered.
	
			filters (:obj:`pypoligram.filters`, *optional*):
				The filter to filter the clients that will receive the handler, defaults to :obj:`pypoligram.filters.ALL`
	
			group (``int``, *optional*):
				The group identifier, defaults to 0.
    
		Returns:
			``tuple``: A tuple consisting of *(handler, group)*.
   
		Example:
			.. code-block:: python
	
				from pyrogram import Client
				from pyrogram.handlers import MessageHandler
				from pypoligram import ClientManager, filters as pfilters
	
				async def hello(client, message):
					print(message)
	
				manager = ClientManager([
					Client("my_account1"),
					Client("my_account2"),
				])
	
				manager.add_handler(MessageHandler(hello), pfilters.client("my_account1"), group=1)
	
				manager.run()
		"""
		names = []
		for client in self:
			if filters(self, client):
				client.add_handler(handler, group)
				names.append(client.name)
		if isinstance(kwargs.get("module_path"), str):
			log.info('[{}] [MULTILOAD] {}("{}") in group {} from "{}" added to {}'.format(
				self.name, type(handler).__name__, kwargs.get("name"), group, kwargs.get("module_path"), (", ".join(repr(name) for name in names[:-1])+f" and {names[-1]!r}") if len(names) > 1 else repr(names[0])))
		return handler, group

	def remove_handler(self, handler: Handler, group: int = 0, **kwargs) -> None:
		"""Remove a previously-registered update handler from multiple clients.
  
		Make sure to provide the right group where the handler was added in. You can use the return value of the
		:meth:`~pyrogram.Client.add_handler` method, a tuple of *(handler, group)*, and pass it directly.
  
		Parameters:
			handler (``Handler``):
				The handler to be removed.
  
			group (``int``, *optional*):
				The group identifier, defaults to 0.
    
		Example:
			.. code-block:: python
	
				from pyrogram import Client
				from pyrogram.handlers import MessageHandler
				from pypoligram import ClientManager
	
				async def hello(client, message):
					print(message)
	
				manager = ClientManager([
					Client("my_account1"),
					Client("my_account2"),
				])
	
				handler = manager.add_handler(MessageHandler(hello), pfilters.client("my_account1"))
	
				# Starred expression to unpack (handler, group)
				manager.remove_handler(*handler)
	
				manager.run()
		"""
		names = []
		for client in self:
			try:
				client.remove_handler(handler, group)
				names.append(client.name)
			except ValueError:
				pass
		if isinstance(kwargs.get("module_path"), str):
			log.info('[{}] [MULTIUNLOAD] {}("{}") in group {} from "{}" removed from {}'.format(
				self.name, type(handler).__name__, kwargs.get("name"), group, kwargs.get("module_path"), (", ".join(repr(name) for name in names[:-1])+f" and {names[-1]!r}") if len(names) > 1 else repr(names[0])))

	def load_plugins(self):
		if self.plugins:
			plugins = self.plugins.copy()

			for option in ["include", "exclude"]:
				if plugins.get(option, []):
					plugins[option] = [
						(i.split()[0], i.split()[1:] or None)
						for i in self.plugins[option]
					]
		else:
			return

		if plugins.get("enabled", True):
			root = plugins["root"]
			include = plugins.get("include", [])
			exclude = plugins.get("exclude", [])

			count = 0

			if not include:
				for path in sorted(Path(root.replace(".", "/")).rglob("*.py")):
					module_path = '.'.join(path.parent.parts + (path.stem,))
					module = import_module(module_path)

					for name in vars(module).keys():
						# noinspection PyBroadException
						try:
							for handler, filters, group in getattr(module, name).handlers:
								if isinstance(handler, Handler) and isinstance(group, int):
									self.add_handler(handler, filters, group, name=name, module_path=module_path)

									# log.info('[{}] [MULTILOAD] {}("{}") in group {} from "{}"'.format(
									#	self.name, type(handler).__name__, name, group, module_path))

									count += 1
						except Exception:
							pass
			else:
				for path, handlers in include:
					module_path = root + "." + path
					warn_non_existent_functions = True

					try:
						module = import_module(module_path)
					except ImportError:
						log.warning('[%s] [MULTILOAD] Ignoring non-existent module "%s"', self.name, module_path)
						continue

					if "__path__" in dir(module):
						log.warning('[%s] [MULTILOAD] Ignoring namespace "%s"', self.name, module_path)
						continue

					if handlers is None:
						handlers = vars(module).keys()
						warn_non_existent_functions = False

					for name in handlers:
						# noinspection PyBroadException
						try:
							for filters, handler, group in getattr(module, name).handlers:
								if isinstance(handler, Handler) and isinstance(group, int):
									self.add_handler(filters, handler, group, name=name, module_path=module_path)

									# log.info('[{}] [MULTILOAD] {}("{}") in group {} from "{}"'.format(
									#	self.name, type(handler).__name__, name, group, module_path))

									count += 1
						except Exception:
							if warn_non_existent_functions:
								log.warning(f'[{self.name}] [MULTILOAD] Ignoring non-existent function "{name}" from "{module_path}"')

			if exclude:
				for path, handlers in exclude:
					module_path = root + "." + path
					warn_non_existent_functions = True

					try:
						module = import_module(module_path)
					except ImportError:
						log.warning('[%s] [MULTIUNLOAD] Ignoring non-existent module "%s"', self.name, module_path)
						continue

					if "__path__" in dir(module):
						log.warning('[%s] [MULTIUNLOAD] Ignoring namespace "%s"', self.name, module_path)
						continue

					if handlers is None:
						handlers = vars(module).keys()
						warn_non_existent_functions = False

					for name in handlers:
						# noinspection PyBroadException
						try:
							for filters, handler, group in getattr(module, name).handlers:
								if isinstance(handler, Handler) and isinstance(group, int):
									self.remove_handler(filters, handler, group, name=name, module_path=module_path)

									# log.info('[{}] [MULTIUNLOAD] {}("{}") from group {} in "{}"'.format(
									#	self.name, type(handler).__name__, name, group, module_path))

									count -= 1
						except Exception:
							if warn_non_existent_functions:
								log.warning(f'[{self.name}] [MULTIUNLOAD] Ignoring non-existent function "{name}" from "{module_path}"')

			if count > 0:
				log.info('[{}] Successfully loaded {} plugin{} from "{}"'.format(
					self.name, count, "s" if count > 1 else "", root))
			else:
				log.warning('[%s] No plugin loaded from "%s"', self.name, root)

	def __iter__(self) -> Self:
		self.__iter: Iterable[Client] = iter(self._clients)
		return self

	async def __aiter__(self) -> Self:
		self.__aiter: Iterable[Client] = iter(self._clients)
		return self

	def __next__(self) -> Client:
		return next(self.__iter)

	async def __anext__(self) -> Client:
		try:
			return next(self.__aiter)
		except StopIteration:
			raise StopAsyncIteration

	def __enter__(self):
		return self.start()

	def __exit__(self, *args):
		try:
			self.stop()
		except ConnectionError:
			pass

	async def __aenter__(self):
		return await self.start()

	async def __aexit__(self, *args):
		try:
			await self.stop()
		except ConnectionError:
			pass

	async def start(self, sequential: bool = False) -> Self:
		"""Start all clients in the manager.
		
  		You can use it just like :meth:`pyrogram.Client.start` but it will start all clients in the manager.
    
		Parameters:
			sequential (``bool``, *optional*):
				Start the clients sequentially instead of concurrently.
				Defaults to False.
    
		Returns:
			:obj:`~pypoligram.ClientManager`: The started manager itself.
   
		Raises:
			ConnectionError: In case you try to start an already started client.
   
		Example:
			.. code-block:: python
   
				import asyncio
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])

	
				async def main():
					await manager.start()
					... # Invoke API methods
					await manager.stop()
	
				asyncio.run(main())
		"""
		self.load_plugins()
		if sequential:
			for client in self:
				await client.start()
			return self
		await asyncio.gather(*[client.start() for client in self])
		return self

	async def stop(self, sequential: bool = False, block: bool = True) -> Self:
		"""Stop all clients in the manager.
		
  		You can use it just like :meth:`pyrogram.Client.stop` but it will stop all clients in the manager.
    
		Parameters:
			sequential (``bool``, *optional*):
				Stop the clients sequentially instead of concurrently.
				Defaults to False.
	
			block (``bool``, *optional*):
				Blocks the code execution until all the clients has been stopped. It is useful with ``block=False`` in
				case you want to stop the own client *within* a handler in order not to cause a deadlock. 
				Defaults to True.
    
		Returns:
			:obj:`~pypoligram.ClientManager`: The stopped manager itself.
   
		Raises:
			ConnectionError: In case you try to stop an already stopped client.
   
		Example:
			.. code-block:: python
	
				import asyncio
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])
	
				async def main():
					await manager.start()
					... # Invoke API methods
					await manager.stop()
     
				asyncio.run(main())
		"""
		async def do_it():
			if sequential:
				for client in self:
					await client.stop()
				return self
			await asyncio.gather(*[client.stop() for client in self])
		if block:
			await do_it()
		else:
			self.loop.create_task(do_it())

		return self

	async def restart(self, sequential: bool = False, block: bool = True) -> Self:
		"""Restart all clients in the manager.
		
  		This method calls :meth:`pyrogram.Client.restart` for every client in the manager.
    
		Parameters:
			sequential (``bool``, *optional*):
				Restart the clients sequentially instead of concurrently.
				Defaults to False.
    
			block (``bool``, *optional*):
				Blocks the code execution until all the clients has been restarted. It is useful with ``block=False`` in
				case you want to restart the own client *within* a handler in order not to cause a deadlock. 
				Defaults to True.
    
		Returns:
			:obj:`~pypoligram.ClientManager`: The restarted manager itself.
   
		Raises:
			ConnectionError: In case you try to restart a stopped client.
   
		Example:
			.. code-block:: python
	
				import asyncio
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])
	
				async def main():
					await manager.start()
					... # Invoke API methods
					await manager.restart()
					... # Invoke other API methods
					await manager.stop()
     
				asyncio.run(main())
		"""
		async def do_it():
			if sequential:
				for client in self:
					await client.restart()
				return self
			await asyncio.gather(*[client.restart() for client in self])

		if block:
			await do_it()
		else:
			self.loop.create_task(do_it())

		return self

	async def restart2(self, sequential: bool = False, block: bool = True) -> Self:
		"""Restart all clients in the manager.
		
  		This method will call :meth:`pypoligram.ClientManager.stop` and then :meth:`pypoligram.ClientManager.start` in a row instead of
		calling :meth:`pyrogram.Client.restart` for every client in the manager.
  
		Parameters:
			sequential (``bool``, *optional*):
				Restart the clients sequentially instead of concurrently.
				Defaults to False.
	
			block (``bool``, *optional*):
				Blocks the code execution until all the clients has been restarted. It is useful with ``block=False`` in
				case you want to restart the own client *within* a handler in order not to cause a deadlock. 
				Defaults to True.
    
		Returns:
			:obj:`~pypoligram.ClientManager`: The restarted manager itself.
   
		Raises:
			ConnectionError: In case you try to restart a stopped client.
   
		Example:
			.. code-block:: python
	
				import asyncio
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])
	
				async def main():
					await manager.start()
					... # Invoke API methods
					await manager.restart()
					... # Invoke other API methods
					await manager.stop()
     
				asyncio.run(main())
		"""
		async def do_it():
			await self.stop(sequential=sequential)
			await self.start(sequential=sequential)

		if block:
			await do_it()
		else:
			self.loop.create_task(do_it())

		return self

	def run(self, coroutine=None, /, *, sequential: bool = False) -> None:
		"""Start the manager, idle the main script and finally stop the manager.
  
		When calling this method without any argument it acts as a convenience method that calls
		:meth:`~pypoligram.ClientManager.start`, :meth:`~pyrogram.idle` and :meth:`~pypoligram.ClientManager.stop` in sequence.
		It makes running a single client less verbose.
  
		In case a coroutine is passed, runs the coroutine until it's completed and doesn't do any client
        operation. This is almost the same as :py:obj:`asyncio.run` except for the fact that PyPoligram's ``run`` uses the
        current event loop instead of a new one.
        
		Parameters:
			coroutine (``coroutine``, *optional*):
				The coroutine to run. If not provided, the manager will be started and idled.
	
			sequential (``bool``, *optional*):
				Start the clients sequentially instead of concurrently.
				Defaults to False.
    
		Raises:
			ConnectionError: In case you try to run an already started client.
   
		Example:
			.. code-block:: python
	
				from pyrogram import Client
				from pypoligram import ClientManager
	
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])
				... # Set handlers up
				manager.run()
	
			.. code-block:: python
	
				from pyrogram import Client
				from pypoligram import ClientManager
				
				manager = ClientManager([
					Client("account1"),
					Client("account2"),
					Client("account3")
				])
    
				async def main():
					async with manager:
						... # Do something with the clients
      
      
				manager.run(main())
		"""
		loop = asyncio.get_event_loop()
		run = loop.run_until_complete
		from pyrogram.methods.utilities.idle import idle

		if coroutine is not None:
			run(coroutine)
		elif inspect.iscoroutinefunction(self.start):
			run(self.start(sequential))
			run(idle())
			run(self.stop(sequential=sequential))
		else:
			self.start(sequential)
			run(idle())
			self.stop(sequential=sequential)



import inspect
import logging

import pyrogram
from pyrogram import dispatcher
from pyrogram.handlers import RawUpdateHandler

import pypoligram

log = logging.getLogger(__name__)

class Dispatcher(dispatcher.Dispatcher):
	def __init__(self, client: "pyrogram.Client", clients: "pypoligram.ClientManager") -> None:
		self.clients = clients
		super().__init__(client)

	async def handler_worker(self, lock):
		while True:
			packet = await self.updates_queue.get()

			if packet is None:
				break

			try:
				update, users, chats = packet
				parser = self.update_parsers.get(type(update), None)

				parsed_update, handler_type = (
					await parser(update, users, chats)
					if parser is not None
					else (None, type(None))
				)

				async with lock:
					for group in self.groups.values():
						for handler in group:
							args = None

							if isinstance(handler, handler_type):
								try:
									if await handler.check(self.client, parsed_update):
										args = (parsed_update,)
								except Exception as e:
									log.exception(e)
									continue

							elif isinstance(handler, RawUpdateHandler):
								args = (update, users, chats)

							if args is None:
								continue

							try:
								if inspect.iscoroutinefunction(handler.callback):
									await handler.callback(self.clients, self.client, *args)
								else:
									await self.loop.run_in_executor(
										self.client.executor,
										handler.callback,
										self.clients,
										self.client,
										*args
									)
							except pyrogram.StopPropagation:
								raise
							except pyrogram.ContinuePropagation:
								continue
							except Exception as e:
								log.exception(e)

							break
			except pyrogram.StopPropagation:
				pass
			except Exception as e:
				log.exception(e)
